// 
// Module: RmfTableIterator.h
// Author: Hassan Siddiqui
//
// Description: RMF generation task for the XMM Science Analysis System
// 
// Comments: 
//  This prototype is restricted the central CCD of the MOS1 detector.
//  
// Modifications: See ChangeLog
// 
// $Id: RmfTableIterator.cc,v 1.5 2000/11/01 09:16:58 hsiddiqu Exp $

/* General classes */
#include <iostream>

/* SAS-specific classes */
#include "Dal.h"
#include "FileLocator.h"
#include "Tracer.h"
#include "DumpVector.h"
#include "NumericalRange.h"

/* Local classes */
#include "RmfTableIterator.h"
#include "RmfDataServer.h"
#include "ResponseGrouping.h"
#include "ResponseRow.h"
#include "RmfUtils.h"
#include "MatrixColumn.h"

/***********************************************
 *
 * class RmfTableIterator
 *
 * Table interator class
 *
 ***********************************************/

/* Constructors */

RmfTableIterator::RmfTableIterator( RmfDataServer* rmfDataServer,
				    double threshold
				    ) :
  _rmfDataServer(rmfDataServer),
  _threshold(threshold),
  _ebinsloColData(0), 
  _ebinshiColData(0), 
  _n_grpColData(0),
  _f_chanColData(0),
  _n_chanColData(0),
  _matrixColumn(0), 
  _max_ngroups(0),
  _max_nresponse_elements(0)
{
  PROC;

  if (_threshold <= FLT64_MIN) _threshold = 0;
  
  getGrouping();
  
  ENDPROC;
}

RmfTableIterator::~RmfTableIterator()
{
  errstr << " ~RmfTableIterator() called."
	 << Sas::message(Msg::App, Msg::Verbose);
  
  // Delete ColumnData objects
  delete _ebinsloColData;
  _ebinsloColData = 0;
  
  delete _ebinshiColData;
  _ebinshiColData = 0;

  delete _n_grpColData;
  _n_grpColData = 0;

  delete _f_chanColData;
  _f_chanColData = 0;

  delete _n_chanColData;
  _n_chanColData = 0;

  delete _matrixColumn;
  _matrixColumn = 0;

/* Remove the mega-array */
  //delete _responseMatrix;
  //_responseMatrix = 0;
}

/* Initialise Table */
void RmfTableIterator::table(Table*)
    {
      PROC;

      // Don't do anything until subtable iteration implemented 
      // (and SubTable::addColumn method complete)
      // use temporary setupTable() method.

      ENDPROC;
    }

void RmfTableIterator::setupTable(Table* table )
    {
      PROC;
      
      /* NOTE: There are some FTOOLs which require the columns to be
	 arranged in a specific order. Eg chkrmf requires the F_CHAN column
	 to be 4th in the table. 
	 Will add columns in the order specified in OGIP 92/002.
      */


      /* Create ENERG_LO and ENERG_HI columns and set up references to them */

      Column* ebinsloCol = table->addColumn("ENERG_LO", Column::Real32, "No comment", "keV");

      _ebinsloColData = ebinsloCol->data();

      Column* ebinshiCol = table->addColumn("ENERG_HI", Column::Real32, "No comment", "keV");

      _ebinshiColData = ebinshiCol->data();


      /* N_GRP */

      Column* n_grpCol  = table->addColumn("N_GRP",  Column::Int16, "No comment", "");

      _n_grpColData = n_grpCol->data();


      /* F_CHAN */

      vector<unsigned long> sizeVector_fchan;

      sizeVector_fchan.push_back(_max_ngroups);

      SAS_DEBUG( cout << " size of fchan " << _max_ngroups <<endl;);

      Column* f_chanCol = table->addColumn("F_CHAN", 
					   Column::Int16, 
					   "No comment", 
					   "", 
					   sizeVector_fchan);


      // Check that this is the 4th column.
      // assert(table->columnNumber("F_CHAN") == 3);
     


      _f_chanColData = f_chanCol->data();


      // Add attribute information to this column.
      // This satisfies the FTOOL CHKRMF.

      *(f_chanCol->addAttribute("TLMIN", "Minimum value legally allowed in FCHAN column")) = (int)_rmfDataServer->tlmin();

      *(f_chanCol->addAttribute("TLMAX", "Minimum value legally allowed in FCHAN column")) = (int)_rmfDataServer->tlmax();


      /* N_CHAN */

      vector<unsigned long> sizeVector_nchan;

      sizeVector_nchan.push_back(_max_ngroups);

      Column* n_chanCol = table->addColumn("N_CHAN", 
				       Column::Int16, 
				       "No comment", 
				       "", 
				       sizeVector_nchan);

      _n_chanColData = n_chanCol->data();


      /* MATRIX */

      vector<unsigned long> sizeVector_matrix;

      sizeVector_matrix.push_back(_max_nresponse_elements);

      errstr << " Number of elements in a row of matrix " 
	     << _max_nresponse_elements 
	     << Sas::message(Msg::App, Msg::Verbose); 
      
      _matrixColumn =  _rmfDataServer->matrixColumn(table, sizeVector_matrix);

      ENDPROC;
    }


  // -----------------------------------------

  // Get grouping information

  // -----------------------------------------

void RmfTableIterator::getGrouping()
    {
      PROC;

      // This function determines how the output response matrix will be grouped.

      // Go through matrix to find out the following statistics: 
      // 1. The largest number of pha groups in a row
      // 2. The largest number of non-zero pha elements in a row

      unsigned long row_max_nresponse_elements = 0;
      unsigned long row_max_ngroups = 0;

      unsigned long ngroups;
      unsigned long nresponse_elements;

      for (unsigned int row = 
	     _rmfDataServer->response()->ebinFirstIndex(); 
	   row <= _rmfDataServer->response()->ebinLastIndex(); 
	   ++row)
	{
         /* Get the response for this energy/row */
          const vector<double>& response = _rmfDataServer->
                         response()->responseVector(row);

         /* Add this into the full matrix */
          //(*_responseMatrix).push_back(response); 
          _responseMatrix.push_back(response); 
	  
	  // ** Compute statistics :

	  // Create a new object to compute 
	  // the grouping information for the given row.
	  ResponseRow response_row(response, _threshold);


	  // Determine row with largest number of groups

	  ngroups = response_row.ngroups();

	  if (ngroups > _max_ngroups) 
	    {
	      _max_ngroups = ngroups;
	      row_max_ngroups = row;
	    }

	  // Determine row with largest number of elements above threshold

	  nresponse_elements = response_row.nresponse_elements();

	  if (nresponse_elements > _max_nresponse_elements)
	    {
	      _max_nresponse_elements = nresponse_elements;
	      row_max_nresponse_elements = row;
	    }

	}

      errstr << " largest number of pha elements in response above " << _threshold 
	     << " is " << _max_nresponse_elements 
	     << " found in row " << row_max_nresponse_elements
	     << endl
	     << " the largest number of groups in response "
	     << " is " << _max_ngroups 
	     << " found in row " << row_max_ngroups
	     << " for a threshold of " << _threshold
	     << Sas::message(Msg::App, Msg::Verbose);

      if (!_max_nresponse_elements) 
	{
	  Sas::error ("noPhaElements", 
		 "There are no pha elements in the response above the threshold value.");
	}

      if (!_max_ngroups) 
	{
	  errstr << "There are " << _max_nresponse_elements
		 << " in row " << row_max_nresponse_elements
		 << " but there are no groups found. Please raise an SPR."
		 << Sas::error ("groupingInvalid"); 
	}

      ENDPROC;
    }


  // ---------------------------------------------------

  // Subtable iteration function

  // ---------------------------------------------------

//   void proc(SubTable& ) 
//     {
//       PROC;

//       // Don't do any subtable iterations until from() method defined.
//       // Hence, make proc() redundant and 
//       // setup a temporary processEntireTable() method.

//       ENDPROC
//     } 


void RmfTableIterator::processEntireTable(Table* table)
    {
      PROC;

      setupTable(table);


      // Get pointers to actual data

      real32* ebinsloData = _ebinsloColData ->real32Data();

      real32* ebinshiData = _ebinshiColData ->real32Data();


      // Populate columns.

      errstr << " populating matrix columns ... " 
	     << Sas::message(Msg::App, Msg::Verbose);

      unsigned int ebinIndex = _rmfDataServer->response()->ebinFirstIndex();

      for (unsigned long row = 0; row < table->rows(); row++)
	{
	    if (!(row%100)) 
	    {
	      errstr <<"RmfTableIterator: populating row number " << row 
		     << Sas::message(Msg::App, Msg::Verbose);
	    }

	  // Get energy bins
	  Interval<double> ebin = _rmfDataServer->response()->ebins(ebinIndex);

	  // Add entries to ENERG_LO and ENERG_HI columns

	  /* NOTE: energy ranges in eV - convert to keV */

	  ebinsloData[row] = ebin.lo()/1000.;

	  ebinshiData[row] = ebin.hi()/1000.;

	      errstr <<"RmfTableIterator: energy range: "<<ebinsloData[row]
                     <<" "<<ebinshiData[row]
		     << Sas::message(Msg::App, Msg::Verbose);

	  // Get response for given row

	  // UNCOMMENT AFTER DEBUGGING
// 	  const vector<double>& response = getResponse(ebin);

 	  //const vector<double>& response = 
	   // _rmfDataServer->response()->responseVector(ebinIndex);
          //const vector<double>& response = (* _responseMatrix)[ebinIndex];
          const vector<double>& response = _responseMatrix[ebinIndex];

	  // ** Populate matrix column

	  // Group row

	  ResponseRow responserow(response, _threshold);

	  // Write grouped row to column

	  responserow.writeTo(row, 
			      _matrixColumn, 
			      _n_grpColData, 
			      _f_chanColData, 
			      _n_chanColData,
                              _rmfDataServer->tlmin());
	  ebinIndex++;

	  ENDPROC;
	}

    }  


/* Get row response for specified energy range */
const vector<double> RmfTableIterator::getResponse(const Interval<double>& ebin)
    {
      PROC;

      // Return corrected response

      vector<double> correctedresponse = 
	RmfUtils::normalizeResponse(_rmfDataServer->
				    responseCorrector()->
				    correctedResponse( ebin ));
      
      RETURN correctedresponse; 

    }
