// Module: VarMatrixColumn
// Author: Hassan Siddiqui
//
// Description:
// Variable length vector column version of MatrixColumn
// 
// Comments: 
//  
// Modifications: See ChangeLog
// 
// $Id: VarMatrixColumn.cc,v 1.1 2000/03/07 13:49:44 hsiddiqu Exp $

/* General */

/* SAS-specific */
#include "Dal.h"
#include "errstr.h"
#include "Tracer.h"

/* Local */
#include "VarMatrixColumn.h"

/* Constructor */
VarMatrixColumn::VarMatrixColumn( Table * table ):
  _matrixCol(0),
  _cellData(0)
{
  PROC;

  /* Create new Column */
  _matrixCol = table->addColumn("MATRIX", 
				Column::Real32, 
				"No comment", 
				"", 
				Dal::zero);
  errstr << " VarMatrixColumn object created. "
	 << Sas::message(Msg::App, Msg::Verbose);

  ENDPROC;
}


/* Destructor */
VarMatrixColumn::~VarMatrixColumn()
{
  PROC;
  errstr << " Destroying VarMatrixColumn object ... "
	 << Sas::message(Msg::App, Msg::Verbose);

  if (_cellData)
    {
      errstr << " cellData object for VarMatrixColumn is still set before destructor was accessed. This should not occur, please raise an SPR."
	     << Sas::error("invalidCellData");
      ALTRETURN;
    }

  ENDPROC;
}

void VarMatrixColumn::row( unsigned long row,
			   unsigned int  size)
{
  PROC;

  if (_cellData)
    {
      errstr << " cellData object for VarMatrixColumn is still set before row() method was accessed. This should not occur, please raise an SPR."
	     << Sas::error("invalidCellData");
      ALTRETURN;
    }

  _cellData = _matrixCol->cellData(row);

  errstr << "VarMatrixColumn: setting size of cell in row " << row
	 << " to " << size
	 << Sas::message(Msg::App, Msg::Verbose);
  _cellData->size(size);

  errstr << "VarMatrixColumn: initialising cell with zeroes"
	 << Sas::message(Msg::App, Msg::Verbose);
  for (unsigned int element = 0; element < _cellData->size(); ++element)
    {
      _cellData-> real32Data()[ element ] = 0;
    }

  _row = row;
  _rowelements = size;

  ENDPROC;
}

unsigned int VarMatrixColumn::rowElements() const
{
  PROC;
  RETURN _rowelements;
}

/* Write to a given position in column */
void VarMatrixColumn::writeTo( unsigned int element,
			       double value)
{
  PROC;

  if ( element > _rowelements)
    {
      errstr << " element specified [" << element
	     << "] is out of bounds for matrix column [0:" << _rowelements-1
	     << Sas::error("elementOutOfBounds");
      ALTRETURN;
    }

  if (!_cellData)
    {
      errstr << " cellData object for VarMatrixColumn is not set in writeTo() method. This should not occur, please raise an SPR."
	     << Sas::error("invalidCellData");
      ALTRETURN;
    }

  _cellData-> real32Data()[ element ] = value;
 
  errstr << "matrix ["
	 << _row
	 << ","
	 << element
	 << "] set to "
	 << _cellData-> real32Data()[ element ]
	 << Sas::message(Msg::App, Msg::Noisy);

  ENDPROC;
}

void VarMatrixColumn::closeRow()
{
  if (!_cellData)
    {
      errstr << " cellData object for VarMatrixColumn is not set in closeRow() method. This should not occur, please raise an SPR."
	     << Sas::error("invalidCellData");
      ALTRETURN;
    }
  
  errstr << "closeRow(): deleting cellData object .."
	 << Sas::message(Msg::App, Msg::Verbose);

  delete _cellData; _cellData = 0;
}
